/*----------------------------------------------------------------------------
 * Standard C library headers
 *----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>
#include <math.h>
#include <string.h>
#include <float.h>
#include <unistd.h>

/*----------------------------------------------------------------------------
 *  Header for the current file
 *----------------------------------------------------------------------------*/

#include "pdm_part_comm_graph.h"
#include "pdm_part_comm_graph_priv.h"
#include "pdm.h"
#include "pdm_timer.h"
#include "pdm_priv.h"
#include "pdm_binary_search.h"
#include "pdm_sort.h"
#include "pdm_printf.h"
#include "pdm_error.h"
#include "pdm_array.h"
#include "pdm_order.h"
#include "pdm_logging.h"
#include "pdm_distrib.h"
#include "pdm_unique.h"

#ifdef __cplusplus
extern "C" {
#if 0
} /* Fake brace to force back Emacs auto-indentation back to column 0 */
#endif
#endif /* __cplusplus */



/*=============================================================================
 * Macro definitions
 *============================================================================*/

/*============================================================================
 * Type
 *============================================================================*/

/*=============================================================================
 * Static global variables
 *============================================================================*/


/*=============================================================================
 * Static function definitions
 *============================================================================*/

static
void
_prepare_send_strid_cst
(
 PDM_part_comm_graph_t   *pcg,
 size_t                   s_data,
 int                      cst_stride,
 void                   **send_entity_data,
 unsigned char          **send_buffer
)
{

  int n_rank;
  PDM_MPI_Comm_size(pcg->comm, &n_rank);

  int s_data_tot = s_data * cst_stride;

  int send_buff_size = s_data * cst_stride * pcg->send_idx[n_rank];
  unsigned char *_send_buffer = NULL;
  PDM_malloc(_send_buffer, send_buff_size, unsigned char);

  unsigned char **_send_entity_data = (unsigned char **) send_entity_data;

  for(int i_part = 0; i_part < pcg->n_part; ++i_part) {
    for(int i = 0; i < pcg->n_entity_graph[i_part]; ++i) {
      for(int k = 0; k < s_data_tot; ++k) {
        int idx_read  = i * s_data_tot + k;
        int idx_write = pcg->part_to_send_buffer[i_part][i] * s_data * cst_stride + k;
        _send_buffer[idx_write] = _send_entity_data[i_part][idx_read];
      }
    }
  }
  *send_buffer = _send_buffer;
}


static
void
_post_recv_strid_cst
(
 PDM_part_comm_graph_t   *pcg,
 size_t                   s_data,
 int                      cst_stride,
 unsigned char           *recv_buffer,
 void                  ***recv_entity_data
)
{
  int s_data_tot = s_data * cst_stride;

  unsigned char **_recv_entity_data = NULL;
  PDM_malloc(_recv_entity_data, pcg->n_part, unsigned char *);
  *recv_entity_data = (void **) _recv_entity_data;
  for(int i_part = 0; i_part < pcg->n_part; ++i_part) {
    _recv_entity_data[i_part] = malloc(pcg->n_entity_graph[i_part] * s_data_tot * sizeof(unsigned char));
    for(int i = 0; i < pcg->n_entity_graph[i_part]; ++i) {
      for(int k = 0; k < s_data_tot; ++k) {
        int idx_write = i * s_data_tot + k;
        int idx_read  = pcg->part_to_recv_buffer[i_part][i] * s_data * cst_stride + k;
        _recv_entity_data[i_part][idx_write] = recv_buffer[idx_read];
      }
    }
  }
}

static
void
_exch_strid_cst
(
 PDM_part_comm_graph_t   *pcg,
 size_t                   s_data,
 int                      cst_stride,
 void                   **send_entity_data,
 void                  ***recv_entity_data
)
{
  int n_rank;
  PDM_MPI_Comm_size(pcg->comm, &n_rank);

  int s_data_tot = s_data * cst_stride;

  PDM_MPI_Datatype mpi_type;
  PDM_MPI_Type_create_contiguous(s_data_tot, PDM_MPI_BYTE, &mpi_type);
  PDM_MPI_Type_commit(&mpi_type);

  unsigned char *send_buffer = NULL;
  _prepare_send_strid_cst(pcg,
                          s_data,
                          cst_stride,
                          send_entity_data,
                          &send_buffer);

  int recv_buff_size = s_data * cst_stride * pcg->recv_idx[n_rank];
  unsigned char *recv_buffer = NULL;
  PDM_malloc(recv_buffer, recv_buff_size, unsigned char);
  PDM_MPI_Alltoallv(send_buffer,
                    pcg->send_n,
                    pcg->send_idx,
                    mpi_type,
                    recv_buffer,
                    pcg->recv_n,
                    pcg->recv_idx,
                    mpi_type,
                    pcg->comm);
  PDM_free(send_buffer);

  /* Post-traitement */
  _post_recv_strid_cst(pcg,
                       s_data,
                       cst_stride,
                       recv_buffer,
                       recv_entity_data);

  PDM_free(recv_buffer);

  PDM_MPI_Type_free(&mpi_type);
}


static
void
_exch_strid_var
(
 PDM_part_comm_graph_t   *pcg,
 size_t                   s_data,
 int                    **send_entity_stride,
 void                   **send_entity_data,
 int                   ***recv_entity_stride,
 void                  ***recv_entity_data
)
{
  int n_rank;
  PDM_MPI_Comm_size(pcg->comm, &n_rank);

  int s_data_tot = s_data;

  PDM_MPI_Datatype mpi_type;
  PDM_MPI_Type_create_contiguous(s_data_tot, PDM_MPI_BYTE, &mpi_type);
  PDM_MPI_Type_commit(&mpi_type);

  /* Exchange stride */
  int  *send_stride = NULL;
  PDM_malloc(send_stride, pcg->send_idx[n_rank], int);
  for(int i_part = 0; i_part < pcg->n_part; ++i_part) {
    for(int i = 0; i < pcg->n_entity_graph[i_part]; ++i) {
      int idx_write = pcg->part_to_send_buffer[i_part][i];
      send_stride[idx_write] = send_entity_stride[i_part][i];
    }
  }

  int *recv_stride = NULL;
  PDM_malloc(recv_stride, pcg->recv_idx[n_rank], int);
  PDM_MPI_Alltoallv(send_stride,
                    pcg->send_n,
                    pcg->send_idx,
                    PDM_MPI_INT,
                    recv_stride,
                    pcg->recv_n,
                    pcg->recv_idx,
                    PDM_MPI_INT,
                    pcg->comm);

  if(0 == 1) {
    PDM_log_trace_array_int(send_stride, pcg->send_idx[n_rank], "send_stride ::");
    PDM_log_trace_array_int(recv_stride, pcg->recv_idx[n_rank], "recv_stride ::");
  }

  /* Exchange data */
  int *send_stride_idx = NULL;
  PDM_malloc(send_stride_idx, pcg->send_idx[n_rank]+1, int);
  send_stride_idx[0] = 0;
  for(int i = 0; i < pcg->send_idx[n_rank]; ++i) {
    send_stride_idx[i+1] = send_stride_idx[i] + send_stride[i];
  }

  int *recv_stride_idx = NULL;
  PDM_malloc(recv_stride_idx, pcg->recv_idx[n_rank]+1, int);
  recv_stride_idx[0] = 0;
  for(int i = 0; i < pcg->recv_idx[n_rank]; ++i) {
    recv_stride_idx[i+1] = recv_stride_idx[i] + recv_stride[i];
  }

  int *send_data_idx = NULL;
  PDM_malloc(send_data_idx, n_rank+1, int);
  int *send_data_n   = PDM_array_zeros_int(n_rank);
  send_data_idx[0] = 0;
  for(int i = 0; i < n_rank; ++i) {
    send_data_idx[i+1] = send_data_idx[i];
    for(int j = pcg->send_idx[i]; j < pcg->send_idx[i+1]; ++j) {
      send_data_idx[i+1] += send_stride[j];
      send_data_n  [i  ] += send_stride[j];
      // send_stride[j] = 0;
    }
  }

  int *recv_data_idx = NULL;
  PDM_malloc(recv_data_idx, n_rank+1, int);
  int *recv_data_n   = PDM_array_zeros_int(n_rank);
  recv_data_idx[0] = 0;
  for(int i = 0; i < n_rank; ++i) {
    recv_data_idx[i+1] = recv_data_idx[i];
    for(int j = pcg->recv_idx[i]; j < pcg->recv_idx[i+1]; ++j) {
      recv_data_idx[i+1] += recv_stride[j];
      recv_data_n  [i  ] += recv_stride[j];
    }
  }

  if(0 == 1) {
    PDM_log_trace_array_int(send_data_idx, n_rank+1, "send_data_idx ::");
    PDM_log_trace_array_int(recv_data_idx, n_rank+1, "recv_data_idx ::");
    PDM_log_trace_array_int(send_stride_idx, pcg->send_idx[n_rank]+1, "send_stride_idx ::");
    PDM_log_trace_array_int(recv_stride_idx, pcg->recv_idx[n_rank]+1, "recv_stride_idx ::");
  }

  int send_buff_size = send_data_idx[n_rank] * s_data_tot;
  unsigned char  *send_buffer       = malloc(send_buff_size * sizeof(unsigned char));
  unsigned char **_send_entity_data = (unsigned char **) send_entity_data;

  for(int i_part = 0; i_part < pcg->n_part; ++i_part) {
    int idx_read = 0;
    for(int i = 0; i < pcg->n_entity_graph[i_part]; ++i) {
      for(int j = 0; j < send_entity_stride[i_part][i]; ++j) {
        int idx_buffer = pcg->part_to_send_buffer[i_part][i];
        for(int k = 0; k < s_data_tot; ++k) {
          int idx_write  = (send_stride_idx[idx_buffer] + j) * s_data_tot + k;
          send_buffer[idx_write] = _send_entity_data[i_part][(idx_read+j)*s_data_tot + k];
        }
      }
      idx_read += send_entity_stride[i_part][i];
    }
  }
  PDM_free(send_stride_idx);

  int recv_buff_size = recv_data_idx[n_rank] * s_data_tot;
  unsigned char  *recv_buffer = NULL;
  PDM_malloc(recv_buffer, recv_buff_size, unsigned char);

  PDM_MPI_Alltoallv(send_buffer,
                    send_data_n,
                    send_data_idx,
                    mpi_type,
                    recv_buffer,
                    recv_data_n,
                    recv_data_idx,
                    mpi_type,
                    pcg->comm);
  PDM_free(send_buffer);

  /* Panic verbose */
  // PDM_g_num_t* recv_buffer_dbg = (PDM_g_num_t *) recv_buffer;
  // PDM_log_trace_array_long(recv_buffer_dbg, recv_buff_size/s_data, "recv_buffer_dbg ::");

  /* Post-traitement stride */
  int           **_recv_entity_stride = NULL;
  unsigned char **_recv_entity_data   = NULL;
  PDM_malloc(_recv_entity_stride, pcg->n_part, int           *);
  PDM_malloc(_recv_entity_data  , pcg->n_part, unsigned char *);
  *recv_entity_stride =           _recv_entity_stride;
  *recv_entity_data   = (void **) _recv_entity_data;

  for(int i_part = 0; i_part < pcg->n_part; ++i_part) {
    _recv_entity_stride[i_part] = malloc(pcg->n_entity_graph[i_part] * sizeof(int));
    int recv_buff_size_part = 0;
    for(int i = 0; i < pcg->n_entity_graph[i_part]; ++i) {
      int idx_read  = pcg->part_to_recv_buffer[i_part][i];
      _recv_entity_stride[i_part][i] = recv_stride[idx_read];
      recv_buff_size_part += recv_stride[idx_read];
    }

    PDM_malloc(_recv_entity_data[i_part], recv_buff_size_part * s_data_tot, unsigned char);

    if(1 == 0) {
      PDM_log_trace_array_int(_recv_entity_stride[i_part], pcg->n_entity_graph[i_part], "_recv_entity_stride :");
    }
  }

  /*
   * Post-treatment buffer
   */
  for(int i_part = 0; i_part < pcg->n_part; ++i_part) {
    int idx_write = 0;
    for(int i = 0; i < pcg->n_entity_graph[i_part]; ++i) {
      int idx_buffer = pcg->part_to_recv_buffer[i_part][i];
      for(int j = 0; j < _recv_entity_stride[i_part][i]; ++j) {
        for(int k = 0; k < s_data_tot; ++k) {
          int idx_read  = (recv_stride_idx[idx_buffer] + j) * s_data_tot + k;
          _recv_entity_data[i_part][(idx_write+j)*s_data_tot + k] = recv_buffer[idx_read];
        }
      }
      idx_write += _recv_entity_stride[i_part][i];
    }
  }

  PDM_free(recv_buffer);
  PDM_free(recv_stride_idx);

  PDM_free(send_data_idx);
  PDM_free(recv_data_idx);
  PDM_free(send_data_n);
  PDM_free(recv_data_n);

  PDM_free(send_stride);
  PDM_free(recv_stride);

  PDM_MPI_Type_free(&mpi_type);
}



/*=============================================================================
 * Public function definitions
 *============================================================================*/

PDM_part_comm_graph_t*
PDM_part_comm_graph_create
(
  int            n_part,
  int           *pn_entity_graph,
  int          **pentity_graph,
  PDM_MPI_Comm   comm
)
{
  PDM_part_comm_graph_t *pcg = NULL;
  PDM_malloc(pcg, 1 ,PDM_part_comm_graph_t);

  pcg->comm   = comm;
  pcg->n_part = n_part;

  pcg->n_active_rank_send = 0;
  pcg->n_active_rank_recv = 0;

  int n_rank;
  int i_rank;
  PDM_MPI_Comm_size(comm, &n_rank);
  PDM_MPI_Comm_rank(comm, &i_rank);

  pcg->pentity_graph = pentity_graph;

  int n_g_part = 0;
  PDM_MPI_Allreduce(&n_part, &n_g_part, 1, PDM_MPI_INT, PDM_MPI_SUM, comm);
  pcg->n_g_part = n_g_part;

  PDM_malloc(pcg->part_to_send_buffer, n_part, int *);
  PDM_malloc(pcg->part_to_recv_buffer, n_part, int *);
  PDM_malloc(pcg->n_entity_graph     , n_part, int  );

  PDM_MPI_Datatype mpi_triplet_type;
  PDM_MPI_Type_create_contiguous(3, PDM_MPI_INT, &mpi_triplet_type);
  PDM_MPI_Type_commit(&mpi_triplet_type);

  int *send_n = PDM_array_zeros_int(n_rank);
  int *recv_n = NULL;
  PDM_malloc(recv_n, n_rank, int);
  for (int i_part = 0; i_part < n_part; i_part++) {
    int n_entity_graph = pn_entity_graph[i_part];
    pcg->n_entity_graph[i_part] = n_entity_graph;

    for(int idx_entity = 0; idx_entity < n_entity_graph; ++idx_entity) {
      int t_rank = pentity_graph[i_part][4*idx_entity+1];
      send_n[t_rank]++;
    }
  }

  int *send_idx = NULL;
  PDM_malloc(send_idx, n_rank+1, int);
  send_idx[0] = 0;
  for(int i = 0; i < n_rank; ++i) {
    send_idx[i+1] = send_idx[i] + send_n[i];
    if(send_n[i] > 0) {
      pcg->n_active_rank_send++;
    }
    send_n[i] = 0;
  }
  int *send_triplet = NULL;
  PDM_malloc(send_triplet, 3 * send_idx[n_rank], int);

  for (int i_part = 0; i_part < n_part; i_part++) {
    int n_entity_graph = pn_entity_graph[i_part];

    PDM_malloc(pcg->part_to_send_buffer[i_part], n_entity_graph, int);

    for(int idx_entity = 0; idx_entity < n_entity_graph; ++idx_entity) {
      int t_rank = pentity_graph[i_part][4*idx_entity+1];
      int idx_write = send_idx[t_rank] + send_n[t_rank]++;
      send_triplet[3*idx_write  ] = pentity_graph[i_part][4*idx_entity+2]-1;
      send_triplet[3*idx_write+1] = i_part;
      send_triplet[3*idx_write+2] = pentity_graph[i_part][4*idx_entity+3]-1;

      pcg->part_to_send_buffer[i_part][idx_entity] = idx_write;
    }
  }

  PDM_MPI_Alltoall(send_n, 1, PDM_MPI_INT,
                   recv_n, 1, PDM_MPI_INT, comm);

  int *recv_idx = NULL;
  PDM_malloc(recv_idx, n_rank+1, int);
  recv_idx[0] = 0;
  for(int i = 0; i < n_rank; ++i) {
    recv_idx[i+1] = recv_idx[i] + recv_n[i];
    if(recv_n[i] > 0) {
      pcg->n_active_rank_recv++;
    }
  }

  int *recv_triplet = NULL;
  PDM_malloc(recv_triplet, 3 * recv_idx[n_rank], int);

  if(0 == 1) {
    PDM_log_trace_array_int(send_idx, n_rank+1, "send_idx ::");
    PDM_log_trace_array_int(recv_idx, n_rank+1, "recv_idx ::");
  }

  PDM_MPI_Alltoallv(send_triplet,
                    send_n,
                    send_idx,
                    mpi_triplet_type,
                    recv_triplet,
                    recv_n,
                    recv_idx,
                    mpi_triplet_type,
                    comm);

  /* Maintenant on cherche à retrouver la correspondance */
  int **pentity_indices       = NULL;
  int **pentity_indices_order = NULL;
  PDM_malloc(pentity_indices      , n_part, int *);
  PDM_malloc(pentity_indices_order, n_part, int *);

  for (int i_part = 0; i_part < n_part; i_part++) {
    int n_entity_graph = pn_entity_graph[i_part];
    PDM_malloc(pcg->part_to_recv_buffer[i_part], n_entity_graph, int);

    PDM_malloc(pentity_indices      [i_part], 3 * n_entity_graph, int);
    PDM_malloc(pentity_indices_order[i_part],     n_entity_graph, int);

    for(int idx_entity = 0; idx_entity < n_entity_graph; ++idx_entity) {
      pentity_indices[i_part][3*idx_entity  ] = pentity_graph[i_part][4*idx_entity+1];
      pentity_indices[i_part][3*idx_entity+1] = pentity_graph[i_part][4*idx_entity+2]-1;
      pentity_indices[i_part][3*idx_entity+2] = pentity_graph[i_part][4*idx_entity]-1; // Indices locaux
    }

    PDM_order_lnum_s(pentity_indices[i_part],
                     3,
                     pentity_indices_order[i_part],
                     n_entity_graph);

    PDM_order_array(n_entity_graph,
                    3 * sizeof(int),
                    pentity_indices_order[i_part],
                    pentity_indices      [i_part]);
  }

  /* Post buffer */
  for(int t_rank = 0; t_rank < n_rank; ++t_rank) {
    for(int j = recv_idx[t_rank]; j < recv_idx[t_rank+1]; ++j) {

      int lpart   = recv_triplet[3*j  ];
      int tpart   = recv_triplet[3*j+1];
      int lentity = recv_triplet[3*j+2];
      int n_entity_graph = pn_entity_graph[lpart];

      int to_find[3] = {t_rank, tpart, lentity};

      int pos = PDM_order_binary_search_int(to_find, pentity_indices[lpart], 3, n_entity_graph);
      // if(pos == -1) {
      //   log_trace("Try to find fail = (%i/%i/%i) --> %i \n", t_rank, lpart, lentity, pos);
      // }
      pcg->part_to_recv_buffer[lpart][pentity_indices_order[lpart][pos]] = j;

    }
  }

  if(0 == 1) {
    for(int i_part = 0; i_part < n_part; ++i_part) {
      int n_entity_graph = pn_entity_graph[i_part];
      PDM_log_trace_array_int(pcg->part_to_send_buffer[i_part], n_entity_graph, "pcg->part_to_send_buffer ::");
      PDM_log_trace_array_int(pcg->part_to_recv_buffer[i_part], n_entity_graph, "pcg->part_to_recv_buffer ::");
    }
  }

  for(int i_part = 0; i_part < n_part; ++i_part) {
    PDM_free(pentity_indices      [i_part]);
    PDM_free(pentity_indices_order[i_part]);
  }
  PDM_free(pentity_indices       );
  PDM_free(pentity_indices_order );


  PDM_free(send_triplet);
  PDM_free(recv_triplet);
  PDM_MPI_Type_free(&mpi_triplet_type);

  pcg->send_idx = send_idx;
  pcg->recv_idx = recv_idx;
  pcg->send_n   = send_n;
  pcg->recv_n   = recv_n;

  /* Compute p2p array */
  PDM_malloc(pcg->active_rank_send, pcg->n_active_rank_send, int);
  PDM_malloc(pcg->active_rank_recv, pcg->n_active_rank_recv, int);

  pcg->n_active_rank_send = 0;
  pcg->n_active_rank_recv = 0;

  for(int i = 0; i < n_rank; ++i) {
    if(send_n[i] > 0) {
      pcg->active_rank_send[pcg->n_active_rank_send++] = i;
    }
    if(recv_n[i] > 0) {
      pcg->active_rank_recv[pcg->n_active_rank_recv++] = i;
    }
  }

  /*
   * Compute owner
   */
  pcg->bound_owner = NULL;
  PDM_malloc(pcg->bound_owner, n_part, int *);
  for(int i_part = 0; i_part < n_part; ++i_part) {
    int n_entity_graph = pn_entity_graph[i_part];

    PDM_malloc(pcg->bound_owner[i_part], n_entity_graph, int);
    int *lbound_entity       = NULL;
    PDM_malloc(lbound_entity, n_entity_graph, int);

    for(int idx_entity = 0; idx_entity < n_entity_graph; ++idx_entity) {
      lbound_entity      [idx_entity] = pentity_graph[i_part][4*idx_entity];
      pcg->bound_owner[i_part][idx_entity] = -1;
    }

    int n_unique = PDM_inplace_unique(lbound_entity, 0, n_entity_graph-1);

    int *lowner = PDM_array_const_int(n_unique, -1);
    for(int idx_entity = 0; idx_entity < n_entity_graph; ++idx_entity) {
      int l_entity = pentity_graph[i_part][4*idx_entity];
      int t_rank   = pentity_graph[i_part][4*idx_entity+1];
      int t_part   = pentity_graph[i_part][4*idx_entity+2]-1;
      int pos = PDM_binary_search_int(l_entity, lbound_entity, n_unique);

      pcg->bound_owner[i_part][idx_entity] = pos; // Stockage temporaire

      if(lowner[pos] != 0) {
        if(i_rank < t_rank || (i_rank == t_rank && i_part < t_part)) {
          lowner[pos] = 1;
        } else {
          lowner[pos] = 0;
        }
      }
    }

    /* Last loop to fill */
    for(int idx_entity = 0; idx_entity < n_entity_graph; ++idx_entity) {
      int pos = pcg->bound_owner[i_part][idx_entity];
      pcg->bound_owner[i_part][idx_entity] = lowner[pos];
    }

    PDM_free(lowner);
    PDM_free(lbound_entity);
  }

  return pcg;
}

void
PDM_part_comm_graph_exch
(
 PDM_part_comm_graph_t   *pcg,
 size_t                   s_data,
 PDM_stride_t             t_stride,
 int                      cst_stride,
 int                    **send_entity_stride,
 void                   **send_entity_data,
 int                   ***recv_entity_stride,
 void                  ***recv_entity_data
)
{
  if(t_stride == PDM_STRIDE_CST_INTERLACED) {
    _exch_strid_cst(pcg,
                    s_data,
                    cst_stride,
                    send_entity_data,
                    recv_entity_data);
  } else if (t_stride == PDM_STRIDE_VAR_INTERLACED) {

    _exch_strid_var(pcg,
                    s_data,
                    send_entity_stride,
                    send_entity_data,
                    recv_entity_stride,
                    recv_entity_data);
  } else {
    PDM_error(__FILE__, __LINE__, 0, "PDM_part_comm_graph_exch, wrong t_stride \n");
  }

}

const int*
PDM_part_comm_graph_owner_get
(
 PDM_part_comm_graph_t *pcg,
 int                    i_part
)
{
  return pcg->bound_owner[i_part];
}

void
PDM_part_comm_graph_reorder
(
  PDM_part_comm_graph_t  *pcg,
  int                   **pentity_graph,
  int                   **old_to_new
)
{

  /* Prepare exchange */
  int **send_new_id = NULL;
  PDM_malloc(send_new_id, pcg->n_part, int         *);
  for(int i_part = 0; i_part < pcg->n_part; ++i_part) {
    PDM_malloc(send_new_id[i_part], pcg->n_entity_graph[i_part], int        );
    for(int i = 0; i < pcg->n_entity_graph[i_part]; ++i) {
      int i_entity = pentity_graph[i_part][4*i  ]-1;
      send_new_id[i_part][i] = old_to_new[i_part][i_entity];
    }
  }

  int **recv_new_id = NULL;
  PDM_part_comm_graph_exch(pcg,
                                   sizeof(int),
                                   PDM_STRIDE_CST_INTERLACED,
                                   1,
                                   NULL,
                      (void  **)   send_new_id,
                                   NULL,
                      (void ***)   &recv_new_id);

  for(int i_part = 0; i_part < pcg->n_part; ++i_part) {
    PDM_free(send_new_id[i_part]);
  }
  PDM_free(send_new_id);

  /* Actualisation current and opposite */

  for(int i_part = 0; i_part < pcg->n_part; ++i_part) {
    for(int i = 0; i < pcg->n_entity_graph[i_part]; ++i) {
      int i_entity = pentity_graph[i_part][4*i  ]-1;
      pentity_graph[i_part][4*i  ] = old_to_new[i_part][i_entity]+1; // On suppose que old_to_new commence a 0
      pentity_graph[i_part][4*i+3] = recv_new_id[i_part][i]+1; // On suppose que old_to_new commence a 0
    }
  }

  for(int i_part = 0; i_part < pcg->n_part; ++i_part) {
    PDM_free(recv_new_id[i_part]);
  }
  PDM_free(recv_new_id);
}

void
PDM_part_comm_graph_free
(
 PDM_part_comm_graph_t* pcg
)
{

  for(int i_part = 0; i_part < pcg->n_part; ++i_part) {
    PDM_free(pcg->part_to_send_buffer[i_part]);
    PDM_free(pcg->part_to_recv_buffer[i_part]);
    PDM_free(pcg->bound_owner        [i_part]);
  }
  PDM_free(pcg->part_to_send_buffer);
  PDM_free(pcg->part_to_recv_buffer);
  PDM_free(pcg->n_entity_graph);
  PDM_free(pcg->bound_owner);

  PDM_free(pcg->send_idx);
  PDM_free(pcg->recv_idx);
  PDM_free(pcg->send_n);
  PDM_free(pcg->recv_n);
  PDM_free(pcg->active_rank_send);
  PDM_free(pcg->active_rank_recv);

  PDM_free(pcg);
}




#ifdef __cplusplus
}
#endif /* __cplusplus */
